from typing import get_args, get_origin, Iterable, Union, Literal, Any, TYPE_CHECKING, TypeVar
from abc import ABC, abstractmethod
from contextlib import suppress
from functools import cache

from ..convert import *
from ..aliasing import *
from ..dpi import *
from ..utilities import *
from ..storage import *
from ..messagebox import Messagebox
from ..extensions import extendable
from ..doc import doc_category

import tkinter.ttk as ttk
import tkinter as tk
import json


if TYPE_CHECKING:
    from .window import ObjectEditWindow


T = TypeVar('T')


__all__ = (
    "NewObjectFrameBase",
)


@extendable
@doc_category("Object frames")
class NewObjectFrameBase(ttk.Frame, ABC):
    """
    Base Frame for inside the :class:`ObjectEditWindow` that allows object definition.

    Parameters
    -------------
    class_: Any
        The class we are defining for.
    return_widget: Any
        The widget to insert the ObjectInfo into after saving.
    parent: TopLevel
        The parent window.
    old_data: Any
        The old_data gui data.
    check_parameters: bool
        Check parameters (by creating the real object) upon saving.
        This is ignored if editing a function instead of a class.
    allow_save: bool
        If False, will open in read-only mode.
    """
    origin_window: "ObjectEditWindow" = None

    def __init__(
        self,
        class_: Any,
        return_widget: Union[ComboBoxObjects, ListBoxObjects, None],
        parent = None,
        old_data: Any = None,
        check_parameters: bool = True,
        allow_save = True,
    ):
        self.class_ = class_
        "The type (class) of object being creating"

        self.return_widget = return_widget
        "The widget where the abstract object (ObjectInfo) will be stored"

        self._original_gui_data = None
        self.parent = parent
        self.check_parameters = check_parameters  # At save time
        "Check for parameters on save"

        self.allow_save = allow_save  # Allow save or not allow (eg. viewing SQL data)
        "Saving is allowed"

        self.old_gui_data = old_data  # Set in .load
        "If editing, the value being edited"

        # If return_widget is None, it's a floating display with no return value
        editing_index = None
        if old_data is not None and return_widget is not None:
            current = return_widget.current()
            if current != -1:
                editing_index = current

        self.editing_index = editing_index
        "The index of object being edited"

        super().__init__(master=parent)
        self.init_toolbar_frame(class_)
        self.init_main_frame()

    @staticmethod
    @cache
    def get_cls_name(cls: Any, args: bool = False) -> str:
        """
        Returns the name of the class ``cls`` or
        the original class when the name cannot be obtained.
        If alias exists, alias is returned instead.
        """
        normal_cls = get_origin(cls) or cls
        if (alias := get_aliased_name(normal_cls)) is not None:
            return alias + f" ({name})"
        elif hasattr(normal_cls, "__name__"):
            name = normal_cls.__name__
        elif hasattr(normal_cls, "_name") and normal_cls._name:
            name = normal_cls._name
        else:
            name = str(normal_cls)

        if args and (type_args := get_args(cls)):
            name = (
                name + 
                f"[{ ', '.join([NewObjectFrameBase.get_cls_name(x, True) for x in type_args]) }]"
            )
        
        return name

    @classmethod
    def set_origin_window(cls, window: "ObjectEditWindow"):
        cls.origin_window = window

    @classmethod
    def filter_literals(cls, types_: Iterable):
        "Returns only the literal types from ``types_``"
        return [x for x in types_ if get_origin(x) is Literal]

    @classmethod
    def check_literals(cls, value: str, literal_types: Iterable):
        """
        Checks if the ``value`` is a correct iterable.
        
        Parameters
        -------------
        value: str
            The string value to be checked.
        types: Iterable[Literal]
            An iterable of accepted literals.
        """
        allowed_values = []
        for literal_type in literal_types:
            args = get_args(literal_type)
            if value in args:
                return value

            allowed_values.extend(args)

        raise ValueError(
            f"'{value}' (a string) is not one of accepted types and does not match any literal values.\n"
            f"'Allowed literals: {allowed_values}."
        )

    @classmethod
    def cast_type(cls, value: Any, types: Iterable):
        """
        Tries to convert *value* into one of the types inside *types* (first successful).

        Raises
        ----------
        TypeError
            Could not convert into any type.
        """

        CAST_FUNTIONS = {
            dict: lambda v: convert_to_object_info(json.loads(v))
        }


        types = list(filter(lambda t: t.__module__ == "builtins", types))

        if isinstance(value, str) and str in types:
            return value

        for type_ in types:
            with suppress(Exception):
                cast_funct = CAST_FUNTIONS.get(type_)
                if cast_funct is None:
                    value = type_(value)
                else:
                    value = cast_funct(value)
                break
        else:
            raise TypeError(f"Could not convert '{value}' to any of accepted types.\nAccepted types: '{types}'")

        return value

    def init_main_frame(self):
        frame_main = ttk.Frame(self)
        frame_main.pack(expand=True, fill=tk.BOTH)
        self.frame_main = frame_main

    def init_toolbar_frame(self, class_):
        frame_toolbar = ttk.Frame(self)
        frame_toolbar.pack(fill=tk.X)
        self.frame_toolbar = frame_toolbar

    @property
    def modified(self) -> bool:
        """
        Returns True if the GUI values have been modified.
        """
        current_values = self.get_gui_data()
        return current_values != self._original_gui_data

    def update_window_title(self):
        "Set's the window title according to edit context."
        self.origin_window.title(f"{'New' if self.old_gui_data is None else 'Edit'} {self.get_cls_name(self.class_, True)} object")

    def close_frame(self):
        if self.allow_save and self.modified:
            resp = Messagebox.yesnocancel("Save?", "Do you wish to save?", master=self.origin_window)
            if resp is not None:
                if resp:
                    self.save()
                else:
                    self._cleanup()
        else:
            self._cleanup()

    def new_object_frame(
        self,
        class_,
        widget,
        *args,
        **kwargs
    ):
        """
        Opens up a new object frame on top of the current one.
        Parameters are the same as in :class:`NewObjectFrame` (current class).
        """
        allow_save = kwargs.pop("allow_save", self.allow_save)
        return self.origin_window.open_object_edit_frame(
            class_, widget, allow_save=allow_save, *args, **kwargs
        )

    @abstractmethod
    def to_object(self):
        """
        Creates an object from the GUI data.
        """
        raise NotImplementedError

    @abstractmethod
    def load(self, old_data: Any):
        """
        Loads the old object info data into the GUI.

        Parameters
        -------------
        old_data: Any
            The old gui data to load.
        """
        raise NotImplementedError

    def save(self):
        """
        Save the current object into the return widget and then close this frame.
        """
        try:
            if not self.allow_save or self.return_widget is None:
                raise TypeError("Saving is not allowed in this context!")

            object_ = self.to_object()
            self._update_ret_widget(object_)
            self._cleanup()
        except Exception as exc:
            Messagebox.show_error(
                "Saving error",
                f"Could not save the object.\n{exc}",
                parent=self.origin_window
            )

    def remember_gui_data(self):
        """
        Remembers GUI data for change checking.
        """
        self._original_gui_data = self.get_gui_data()

    @abstractmethod
    def get_gui_data(self) -> Any:
        """
        Returns all GUI values.
        """
        raise NotImplementedError

    def _cleanup(self):
        self.origin_window.clean_object_edit_frame()

    def _update_ret_widget(self, new: Any):
        ind = self.return_widget.count()
        if self.old_gui_data is not None:
            ret_widget = self.return_widget
            if self.editing_index is not None:  # The index of edited item inside return widget
                ind = self.editing_index
                ret_widget.delete(ind)

        self.return_widget.insert(ind, new)
        if isinstance(self.return_widget, ComboBoxObjects):
            self.return_widget.current(ind)
        else:
            self.return_widget.select_clear("0", tk.END)
            self.return_widget.selection_set(ind)



"""
Utility module used for caching.
"""
from typing import Callable
from functools import wraps
import pickle
from .doc import doc_category


__all__ = (
    "cache_result",
)


# Caching
@doc_category("Caching")
def cache_result(max: int = 256):
    """
    Caching function that also allows dictionary and lists to be used as a key.

    Parameters
    --------------
    max: int
        The maximum number of items inside the cache.
    """
    def _decorator(fnc: Callable):
        cache_dict = {}

        @wraps(fnc)
        def wrapper(*args, **kwargs):
            try:
                # Convert to pickle string to allow hashing of non-hashables (dictionaries, lists, ...)
                key = pickle.dumps((*args, kwargs))
            except Exception:
                return fnc(*args, **kwargs)

            result = cache_dict.get(key, Ellipsis)
            if result is not Ellipsis:
                return result

            result = fnc(*args, **kwargs)
            cache_dict[key] = result

            if len(cache_dict) > max:
                items = list(cache_dict.items())[:max // 2]
                cache_dict.clear()
                cache_dict.update(items)

            return result

        return wrapper

    return _decorator



"""
Utility module for documentation.
"""
# Documentation
from typing import Optional, Dict, Literal
from os import environ


__all__ = (
    "doc_category",
    "DOCUMENTATION_MODE",
)

DOCUMENTATION_MODE = bool(environ.get("DOCUMENTATION", False))
if DOCUMENTATION_MODE:
    cat_map: Dict[str, list] = {"Program": {}}


def doc_category(
    cat: str,
    manual: Optional[bool] = False,
    path: Optional[str] = None,
    api_type: Literal["Program", "HTTP"] = "Program"
):
    """
    Used to mark the object for documentation.
    Objects marked with this decorator function will
    have :mod:`sphinx.ext.autodoc` directives generated automatically.

    Parameters
    ------------
    cat: str
        The name of the category to put this in.
    manual: Optional[bool]
        Generate ``function`` directives instead of ``autofunction``.
        Should be used when dealing with overloads.
    path: Optional[str]
        Custom path to the object.
    api_type: Literal["Program", "HTTP"]
        The type of API, the documented item belongs to.
        Defaults to 'Program'
    """
    def _category(item):  # item == class or function
        if DOCUMENTATION_MODE:
            cat_map[api_type][cat].append((item, manual, path))
        return item

    if DOCUMENTATION_MODE:
        if cat not in cat_map[api_type]:
            cat_map[api_type][cat] = []

    return _category



"""
Module used to add extensions to add extensions
to TkClassWizard.
"""
from typing import Callable, TypeVar, Union
from inspect import isclass
from functools import wraps

from .doc import doc_category, DOCUMENTATION_MODE


T = TypeVar('T')


class Extension:
    """
    The extension class.

    Parameters
    ----------
    name: str
        The name of the extension.
    version: str
        The (semantic) version of the extension.
    loader: Callable
        The extension loader, which is called at every extended object initialization / call.
        See :func:`tkclasswiz.extensions.extendable` for more information about how it is called.
    """
    def __init__(self, name: str, version: str, loader: Callable[[T], None]) -> None:
        self._name = name
        self._version = version
        self._loader = loader

    def __repr__(self) -> str:
        return f"Extension(name={self._name}, version={self._version}, loader={self._loader.__name__})"

    def __call__(self, *args: T, **kwargs: T) -> None:
        return self._loader(*args, **kwargs)

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def loader(self):
        return self._loader


@doc_category("Extensions")
def extendable(obj: Union[T, list]) -> T:
    """
    Decorator that makes the obj extendable.

    It wraps the ``obj``, which is a class or a function, into an extension object.
    The extension object will adds 3 methods to the original class or function:

    - ``register_pre_extension``
    - ``register_post_extension``
    - ``get_extensions``
    
    The ``get_extensions`` method just returns the list of registered 
    extensions (:class:`tkclasswiz.extensions.Extension`).

    The ``register_pre_extension`` and ``register_post_extension`` methods allow users to extend
    the functionality of original tkclass wiz classes or functions.
    They accept the extension (:class:`tkclasswiz.extensions.Extension`) parameter.

    Pre-extensions (``register_pre_extension``) get activated / called before the original ``__init__`` method / 
    before the original function and accept the ``loader`` of the extension must accept the same arguments
    as the original ``__init__`` method / original function.

    Post-extensions differ a bit if the thing being extended is a class  or a function.
    They both have in common that they get activated after the original ``__init__`` method call / original function
    call, but they differ in the arguments they receive:

    - In the case of the extended is a class,
      the extension ``loader`` accepts the same arguments as the ``__init__`` method receives.
    - In the case of the extended is a function,
      the extension ``loader`` accepts the same arguments as the original function and an additional parameter,
      which is the result of the original function call. The result parameter is passed to the ``loader`` as the
      last positional argument.


    Parameters
    ---------------
    obj: T
        Function or a class that can be extended.
    """

    if DOCUMENTATION_MODE:
        return obj

    if isclass(obj):
        @wraps(obj, updated=[])
        class ExtendableClass(obj):
            __reg_post_ext__ = []
            __reg_pre_ext__ = []

            def __init__(self, *args, **kwargs):
                for extension in ExtendableClass.__reg_pre_ext__:
                    extension(self, *args, **kwargs)

                super().__init__(*args, **kwargs)

                extension: Extension
                for extension in ExtendableClass.__reg_post_ext__:
                    extension(self, *args, **kwargs)

            @classmethod
            def register_pre_extension(cls, extension: Extension):
                cls.__reg_pre_ext__.append(extension)

            @classmethod
            def register_post_extension(obj, extension: Extension):
                obj.__reg_post_ext__.append(extension)

            @classmethod
            def get_extensions(obj):
                return obj.__reg_pre_ext__, obj.__reg_post_ext__[:]

        return ExtendableClass
    else:
        class ExtendableFunction:
            __reg_post_ext__ = []
            __reg_pre_ext__ = []

            def __init__(self, bind: object = None) -> None:
                self.bind = bind

            def __call__(self, *args, **kwargs):
                if self.bind is not None:
                    extra_args = (self.bind,)  # self reference
                else:
                    extra_args = ()

                for ext in ExtendableFunction.__reg_pre_ext__:
                    ext(*extra_args, *args, **kwargs)

                r = obj(*extra_args, *args, **kwargs)
                
                for ext in ExtendableFunction.__reg_post_ext__:
                    r = ext(*extra_args, *args, r, **kwargs)

                return r

            def __get__(self, instance, cls):
                # Bind the wrapper callable object into a callable object "instance"
                return ExtendableFunction(instance)

            @classmethod
            def register_pre_extension(cls, extension: Extension):
                cls.__reg_pre_ext__.append(extension)

            @classmethod
            def register_post_extension(cls, extension: Extension):
                cls.__reg_post_ext__.append(extension)

            @classmethod
            def get_extensions(obj):
                return obj.__reg_pre_ext__, obj.__reg_post_ext__[:]

        return ExtendableFunction()




from typing import Union, Any

from .frame_base import *
from ..extensions import extendable
from ..doc import doc_category

import tkinter as tk
import tkinter.ttk as ttk


__all__ = (
    "NewObjectFrameNumber",
)


@extendable
@doc_category("Object frames")
class NewObjectFrameNumber(NewObjectFrameBase):
    """
    Frame for use inside the ObjectEditWindow that allows definition of numerical data.

    Parameters
    -----------
    class_: Any
        The class we are defining for.
    return_widget: Any
        The widget to insert the ObjectInfo into after saving.
    parent: TopLevel
        The parent window.
    old_data: Any
        The old_data gui data.
    check_parameters: bool
        Check parameters (by creating the real object) upon saving.
        This is ignored if editing a function instead of a class.
    allow_save: bool
        If False, will open in read-only mode.
    """
    def __init__(
        self,
        class_: Any,
        return_widget: Any,
        parent: NewObjectFrameBase = None,
        old_data: Any = None,
        check_parameters: bool = True,
        allow_save: bool = True
    ):
        super().__init__(class_, return_widget, parent, old_data, check_parameters, allow_save)
        self.storage_widget = ttk.Spinbox(self.frame_main, from_=-9999, to=9999)
        self.storage_widget.pack(fill=tk.X)

        if old_data is not None:  # Edit
            self.load(old_data)

        self.remember_gui_data()

    def load(self, old_data: Union[int, float]):
        self.storage_widget.set(old_data)
        self.old_gui_data = old_data

    def get_gui_data(self) -> Union[int, float]:
        return self.return_widget.get()

    def to_object(self):
        return self.cast_type(self.storage_widget.get(), [self.class_])



"""
Modules contains definitions related to GUI object transformations.
"""
from __future__ import annotations
from typing import Any, Union, List, Generic, TypeVar, Mapping, Optional, Callable
from contextlib import suppress
from inspect import signature
from enum import Enum

from .utilities import import_class
from .extensions import extendable
from .cache import cache_result
from .annotations import *
from .aliasing import *
from .doc import doc_category

import datetime as dt
import warnings
import decimal


__all__ = (
    "ObjectInfo",
    "convert_to_objects",
    "convert_to_object_info",
    "convert_to_dict",
    "convert_from_dict",
    "convert_objects_to_script",
    "_convert_to_objects_cached",
)

TClass = TypeVar("TClass")


CONVERSION_ATTR_TO_PARAM = {
    dt.timezone: {
        "offset": lambda timezone: timezone.utcoffset(None),
        "name": lambda timezone: timezone.tzname(None)
    },
}


@doc_category("Conversion")
def register_object_objectinfo_rule(cls: type, mapping: Optional[dict] = {}, **kwargs):
    """
    Used for adding new conversion rules when converting from Python objects
    into abstract ObjectInfo objects (GUI objects).

    These rules will be used when calling the ``convert_to_object_info`` function.

    If rules for ``cls`` do not exist, the conversion function will assume parameters
    are located under the same attribute name.

    A neat way to avoid registering custom conversion rules, is to store the attributes either under
    the same name as the parameter, or store them under a different name and create a ``@property``
    getter which has the same name as the parameter.

    Parameters
    ------------
    mapping: Optional[Dict[str, Union[Callable[[T], Any]], str]]
        Mapping mapping the parameter name to attribute from which to obtain the value.
        Values of mapping can also be a getter function, that accepts the object as parameter.
    kwargs: Optional[Unpack[str, Union[Callable[[T], Any]], str]]
        Keyword arguments mapping parameter name to attribute names from which to obtain the value.
        Values of mapping can also be a getter function, that accepts the object as parameter.

    Example
    ----------
    .. code-block:: python

        class FILE:
            def __init__(self, filename: str):
                self.fullpath = filename

        register_object_objectinfo_rule(
            FILE,
            filename="fullpath"
        )

        # Timezone
        import datetime
        register_object_objectinfo_rule(
            datetime.timezone,
            offset=lambda timezone: timezone.utcoffset(None)
        )
    """

    if cls not in CONVERSION_ATTR_TO_PARAM:
        CONVERSION_ATTR_TO_PARAM[cls] = {}

    CONVERSION_ATTR_TO_PARAM[cls].update(**kwargs, **mapping)


@doc_category("Conversion")
def get_object_objectinfo_rule_map(cls: type) -> dict:
    """
    Returns the conversion mapping for ``cls``

    Returns
    ----------
    Dict[str, Union[Callable[[T], Any]], str]
        Mapping of parameter name to the attribute name or getter function.
    """
    return CONVERSION_ATTR_TO_PARAM.get(cls, {})


@extendable
@doc_category("Conversion")
class ObjectInfo(Generic[TClass]):
    """
    A GUI object that represents real objects.
    The GUI only knows how to work with ObjectInfo objects, iterables and primitive-like types.

    **NOTE**: If the class has any password like fields, they can be specified by
    the class ``__passwords__`` attribute, which is an iterable of strings.

    Parameters
    -----------------
    class_: type
        Real object's type.
    data: dict
        Dictionary mapping to real object's parameters
    nickname: Optional[str]
        Add a nickname to the defined object for easier
        recognition.
    """
    CHARACTER_LIMIT = 150
    custom_display_map: dict[type, Callable] = {}

    def __init__(
        self,
        class_,
        data: Mapping,
        nickname: Optional[str] = None,
    ) -> None:
        self.class_ = class_
        self.data = data
        self.nickname = nickname
        self.__hash = 0
        self._repr = None

    @extendable
    def __eq__(self, _value: object) -> bool:
        if isinstance(_value, ObjectInfo):
            return self.class_ is _value.class_ and self.data == _value.data

        return False

    def __hash__(self) -> int:
        if not self.__hash:
            try:
                self.__hash = hash(self.data)
            except TypeError:
                self.__hash = -1

        return self.__hash

    def __repr__(self) -> str:
        if self._repr is not None:
            return self._repr

        repr_get = ObjectInfo.custom_display_map.get(self.class_, self._repr_default)
        _ret = repr_get(self)

        self._repr = _ret
        return _ret
    
    @staticmethod
    def _repr_default(object_info: ObjectInfo) -> str:
        _ret: List[str] = []
        if object_info.nickname:
            _ret += f"({object_info.nickname}) "

        name = get_aliased_name(object_info.class_)
        if name is not None:
            name += f'({object_info.class_.__name__})'
        else:
            name = object_info.class_.__name__

        _ret +=  name + "("
        private_params = set()
        if hasattr(object_info.class_, "__passwords__"):
            private_params = private_params.union(object_info.class_.__passwords__)

        for k, v in object_info.data.items():
            if len(_ret) > object_info.CHARACTER_LIMIT:
                break

            if isinstance(v, str):
                if k in private_params:
                    v = len(v) * '*'  # Hide password types

                v = f'"{v}"'
            else:
                v = str(v)

            _ret += f"{k}={v}, "

        _ret: str = ''.join(_ret)
        _ret = _ret.rstrip(", ") + ")"
        if len(_ret) > object_info.CHARACTER_LIMIT:
            _ret = _ret[:object_info.CHARACTER_LIMIT] + "...)"

        return _ret

    @classmethod
    def register_repr(cls, class_: type, repr: Callable[[ObjectInfo], str], inherited: bool = False):
        """
        Registers a custom __repr__ (string representation of object) function.
        The function must as a single parameter accept the ``ObjectInfo`` instance
        being represented as a string.

        Parameters
        ------------
        class_: type
            The class for which this custom ``repr`` is being register.
        repr: Callable[[ObjectInfo], str]
            The function that will provide custom ``__repr__``.
            As a parameter it accepts the ``ObjectInfo`` object.
            It returns a :class:`str` (string).
        inherited: bool
            Boolean flag. Setting it to True will register repr for inherited members as well.
            Defaults to False.
        """
        cls.custom_display_map[class_] = repr
        if inherited:
            for type_ in class_.__subclasses__():
                cls.register_repr(type_, repr, True)

@cache_result(max=1024)
@doc_category("Conversion")
def convert_objects_to_script(object: Union[ObjectInfo, list, tuple, set, str]):
    """
    Converts ObjectInfo objects into equivalent Python code.
    """
    object_data = []
    import_data = []

    if isinstance(object, ObjectInfo):
        object_str = f"{object.class_.__name__}(\n    "
        attr_str = []
        for attr, value in object.data.items():
            value, import_data_ = convert_objects_to_script(value)
            import_data.extend(import_data_)
            attr_str.append(f"{attr}={value},\n")

        import_data.append(f"from {object.class_.__module__} import {object.class_.__name__}")
        object_str += "    ".join(''.join(attr_str).splitlines(True)) + ")"
        object_data.append(object_str)

    elif isinstance(object, (list, tuple, set)):
        _list_data = ["[\n"]
        for element in object:
            object_str, import_data_ = convert_objects_to_script(element)
            _list_data.append(object_str + ",\n")
            import_data.extend(import_data_)

        _list_data = "    ".join(''.join(_list_data).splitlines(keepends=True)) + "]"
        object_data.append(_list_data)

    elif isinstance(object, Enum):
        type_name = type(object).__name__
        import_data.append(f"from {object.__module__} import {type_name}")
        object_name = object._name_
        if not object_name:
            name = f"{type_name}({object._value_})"
        else:
            name = ' | '.join(map(lambda name: f"{type_name}.{name}", object_name.split('|')))

        object_data.append(name)

    else:
        if isinstance(object, str):
            object = object.replace("\\", "\\\\").replace("\n", "\\n").replace('"', '\\"')
            object_data.append(f'"{object}"')
        else:
            object_data.append(str(object))

    return ",".join(object_data).strip(), sorted(set(import_data), key=lambda x: len(x), reverse=True)


@extendable
@cache_result(16_384)
@doc_category("Conversion")
def convert_to_object_info(object_: object, **kwargs):
    """
    Converts an object into ObjectInfo.

    Parameters
    ---------------
    object_: object
        The object to convert.
    """
    def _convert_object_info(object_, object_type, attrs):
        data_conv = {}
        for k, v in attrs.items():
            with suppress(Exception):
                if callable(v):
                    value = v(object_)
                else:
                    value = getattr(object_, v)

                # Check if object is a singleton that is not builtin.
                # Singletons should not be stored as they would get recreated causing issues on save -> load.
                type_value = type(value)
                with suppress(ValueError):  # Some don't have a signature
                    if type_value.__name__ not in __builtins__ and not len(signature(type_value).parameters):
                        continue

                if value is object_:
                    data_conv[k] = value
                else:
                    data_conv[k] = convert_to_object_info(value, **kwargs)

        ret = ObjectInfo(object_type, data_conv)
        return ret

    def get_conversion_map(object_type):
        attrs = CONVERSION_ATTR_TO_PARAM.get(object_type)
        if attrs is None:
            attrs = {k:k for k in get_annotations(object_type)}

        attrs.pop("return", None)
        return attrs

    object_type = type(object_)
    if object_type in {int, float, str, bool, decimal.Decimal, type(None)} or isinstance(object_, Enum):
        if object_type is decimal.Decimal:
            object_ = float(object_)

        return object_

    if isinstance(object_, (set, list, tuple)):
        object_ = [convert_to_object_info(value, **kwargs) for value in object_]
        return object_
    
    if isinstance(object_, dict):
        return ObjectInfo(
            dict,
            {k: convert_to_object_info(v, **kwargs) for k, v in object_.items()},
        )


    attrs = get_conversion_map(object_type)
    return _convert_object_info(object_, object_type, attrs)


@cache_result()
def _convert_to_objects_cached(*args, **kwargs):
    return convert_to_objects(*args, **kwargs)


@doc_category("Conversion")
def convert_to_objects(
    d: Union[ObjectInfo, dict, list],
) -> Union[object, dict, List]:
    """
    Converts :class:`ObjectInfo` instances into actual objects,
    specified by the ObjectInfo.class\ _ attribute.

    Parameters
    -----------------
    d: ObjectInfo | list[ObjectInfo] | dict
        The object(s) to convert. Can be an ObjectInfo object, a list of ObjectInfo objects or a dictionary that is a
        mapping of ObjectInfo parameters.
    cached: bool
        If True items will be returned from cache. ONLY USE FOR IMMUTABLE USE.
    """
    def convert_object_info():
        data_conv = {
            k:
            convert_to_objects(v)
            if isinstance(v, (list, tuple, set, ObjectInfo, dict)) else v
            for k, v in d.data.items()
        }

        new_obj = d.class_(**data_conv)
        return new_obj

    if isinstance(d, (list, tuple, set)):
        return [convert_to_objects(item) for item in d]
    if isinstance(d, ObjectInfo):
        return convert_object_info()
    if isinstance(d, dict):
        return {k: convert_to_objects(v) for k, v in d.items()}

    return d


@cache_result()
@doc_category("Conversion")
def convert_to_dict(d: Union[ObjectInfo, List[ObjectInfo], Any]):
    """
    Converts ObjectInfo into dictionary representation.
    """
    if isinstance(d, ObjectInfo):
        data_conv = {k: convert_to_dict(v) for k, v in d.data.items()}
        return {"type": f"{d.class_.__module__}.{d.class_.__name__}", "data": data_conv, "nickname": d.nickname}

    if isinstance(d, list):
        return [convert_to_dict(x) for x in d]

    if issubclass((type_d := type(d)), Enum):
        return {"type": f"{type_d.__module__}.{type_d.__name__}", "value": d.value}

    return d



@cache_result()
@doc_category("Conversion")
def convert_from_dict(d: Union[dict, List[dict], Any]) -> ObjectInfo:
    """
    Converts previously converted dict back to ObjectInfo.
    """
    if isinstance(d, list):
        result = [convert_from_dict(item) for item in d]
        return result

    if isinstance(d, dict):
        type_ = import_class(d["type"])

        if "value" in d:  # Enum type or a single value type
            return type_(d["value"])

        annotations = get_annotations(type_)
        data = {}
        for k, v in d["data"].items():
            if k in annotations:
                data[k] = convert_from_dict(v)
            else:
                warnings.warn(f"Parameter {k} does not exist in {type_}, ignoring.")

        return ObjectInfo(type_, data, d.get('nickname'))

    return d



"""
Module can be used to mark deprecated classes / parameters / parameter types.
"""
from typing import overload, get_origin
from itertools import chain

from .utilities import issubclass_noexcept
from .annotations import convert_types
from .doc import doc_category


__all__ = (
    "register_deprecated",
    "is_deprecated",
)


@overload
@doc_category("Deprecations", manual=True)
def register_deprecated(cls: type):
    """
    Mark ``cls`` deprecated globally. This function cannot be
    used on built-in Python classes.
    """

@overload
@doc_category("Deprecations", manual=True)
def register_deprecated(cls: type, parameter: str):
    """
    Marks a ``parameter`` to be deprecated under specific ``cls``.
    """

@overload
@doc_category("Deprecations", manual=True)
def register_deprecated(cls: type, parameter: str, *types: type):
    """
    Marks multiple ``types`` to be deprecated for a certain ``parameter`` under ``cls``.
    """

def register_deprecated(cls, parameter: str = None, *types: type):
    if parameter is None:
        cls.__deprecated__ = True

    elif not types:
        cls.__wiz_deprecated_params__ = getattr(cls, "__wiz_deprecated_params__", set())
        cls.__wiz_deprecated_params__.add(parameter)
    else:
        types = tuple(chain.from_iterable(map(convert_types, types)))
        cls.__wiz_deprecated_param_types__ = getattr(cls, "__wiz_deprecated_param_types__", dict())
        cls.__wiz_deprecated_param_types__[parameter] = cls.__wiz_deprecated_param_types__.get(parameter, set())
        cls.__wiz_deprecated_param_types__[parameter].update(types)

@overload
@doc_category("Deprecations", manual=True)
def is_deprecated(cls: type):
    """
    Checks if ``cls`` is deprecated globally.
    """

@overload
@doc_category("Deprecations", manual=True)
def is_deprecated(cls: type, parameter: str):
    """
    Checks if ``parameter`` is deprecated under ``cls``.
    """

@overload
@doc_category("Deprecations", manual=True)
def is_deprecated(cls: type, parameter: str, type_: type):
    """
    Checks if ``type_`` is deprecated for a certain ``parameter`` under ``cls``.
    """

def is_deprecated(cls: type, parameter: str = None, type_: type = None):
    if parameter is None:
        return hasattr(cls, "__deprecated__")

    if type_ is None:
        params = getattr(cls, "__wiz_deprecated_params__", set())
        return parameter in params

    depr_types = getattr(cls, "__wiz_deprecated_param_types__", dict()).get(parameter, set())
    if is_deprecated(type_) or type_ in depr_types:
        return True
    
    origin = get_origin(type_)
    if origin is not None and origin in depr_types:
        return True

    type_ = origin or type_
    for depr_type in depr_types:
        if issubclass_noexcept(type_, depr_type):
            return True

    return False



from typing import get_args, get_origin, Any, List, Literal
from functools import partial
from enum import Enum, Flag

from ..doc import doc_category
from ..deprecation import *
from ..utilities import *
from ..convert import *
from ..dpi import *

from ..messagebox import Messagebox
from ..extensions import extendable
from .tooltip import ListboxTooltip
from .frame_base import *
from ..storage import *

import tkinter as tk
import tkinter.ttk as ttk


__all__ = (
    "NewObjectFrameIterable",
    "NewObjectFrameIterableView",
)


@extendable
@doc_category("Object frames")
class NewObjectFrameIterable(NewObjectFrameBase):
    """
    Frame for use inside the ObjectEditWindow that allows definition of lists / iterables.
    The list annotations should be annotated similarly to the following example: ``list[Union[int, float]]`` 
    or ``Iterable[Union[int, float]]`` or ``Iterable[SomeClass]``


    Parameters
    ------------
    class_: Any
        The class we are defining for.
    return_widget: Any
        The widget to insert the ObjectInfo into after saving.
    parent: TopLevel
        The parent window.
    old_data: Any
        The old_data gui data.
    check_parameters: bool
        Check parameters (by creating the real object) upon saving.
        This is ignored if editing a function instead of a class.
    allow_save: bool
        If False, will open in read-only mode.
    """
    def __new__(cls, *args, **kwargs):
        if kwargs.get("allow_save", True):
            obj = super().__new__(NewObjectFrameIterable)
        else:
            obj = super().__new__(NewObjectFrameIterableView)

        return obj

    def __init__(
        self,
        class_: Any,
        return_widget: Any,
        parent=None,
        old_data: list = None,
        check_parameters: bool = True,
        allow_save = True
    ):
        dpi_5 = dpi_scaled(5)
        super().__init__(class_, return_widget, parent, old_data, check_parameters, allow_save)
        self.storage_widget = w = ListBoxScrolled(self.frame_main, height=20)
        ListboxTooltip(self.storage_widget, 0)

        frame_edit_remove = ttk.Frame(self.frame_main, padding=(dpi_5, 0))
        frame_edit_remove.pack(side="right")
        frame_cp = ttk.Frame(frame_edit_remove)
        frame_cp.pack(fill=tk.X, expand=True, pady=dpi_5)

        ttk.Button(frame_cp, text="Copy", command=w.save_to_clipboard).pack(side="left", fill=tk.X, expand=True)
        ttk.Button(frame_cp, text="Paste", command=w.paste_from_clipboard).pack(side="left", fill=tk.X, expand=True)
        menubtn = ttk.Menubutton(frame_edit_remove, text="Insert item")
        menu = tk.Menu(menubtn)
        menubtn.configure(menu=menu)
        menubtn.pack(fill=tk.X)
        ttk.Button(frame_edit_remove, text="Remove", command=w.delete_selected).pack(fill=tk.X)
        ttk.Button(frame_edit_remove, text="Edit", command=lambda: self._edit_selected()).pack(fill=tk.X)

        frame_up_down = ttk.Frame(frame_edit_remove)
        frame_up_down.pack(fill=tk.X, expand=True, pady=dpi_5)
        ttk.Button(frame_up_down, text="Up", command=lambda: w.move_selection(-1)).pack(side="left", fill=tk.X, expand=True)
        ttk.Button(frame_up_down, text="Down", command=lambda: w.move_selection(1)).pack(side="left", fill=tk.X, expand=True)

        self._list_args = get_args(self.class_)
        args_normal = []
        args_depr = []
        for arg in self._list_args:
            if is_deprecated(arg):
                args_depr.append(arg)
            else:
                args_normal.append(arg)

        self._create_add_menu(args_normal, menu)
        if args_depr:
            menu_depr = tk.Menu()
            self._create_add_menu(args_depr, menu_depr)
            menu.add_cascade(label=">Deprecated", menu=menu_depr)

        w.pack(side="left", fill=tk.BOTH, expand=True)

        if old_data is not None:
            self.load(old_data)

        self.remember_gui_data()

    def _create_add_menu(self, args: list, menu: tk.Menu):
        insert_items = []
        widget = self.storage_widget
        for arg in args:
            if arg is None:
                insert_items.append(...)
                insert_items.append(None)
            elif issubclass_noexcept(arg, Enum) and not issubclass_noexcept(arg, Flag):
                insert_items.append(...)
                insert_items.extend(list(arg))
            elif get_origin(arg) is Literal:
                insert_items.append(...)
                insert_items.extend(get_args(arg))
            else:
                menu.add_command(
                    label=f"New {self.get_cls_name(arg, True)}",
                    command=partial(self.new_object_frame, arg, widget)
                )

        if insert_items:
            menu_insert = tk.Menu(menu)
            for item in insert_items:
                if item is Ellipsis:
                    menu_insert.add_separator()
                else:
                    label = f"'{item}'" if isinstance(item, str) else str(item)
                    menu_insert.add_command(label=label, command=partial(widget.insert, tk.END, item))

            menu.add_cascade(label=">Insert", menu=menu_insert)

    def load(self, old_data: List[Any]):
        self.storage_widget.insert(tk.END, *old_data)
        self.old_gui_data = old_data

    def get_gui_data(self) -> Any:
        return self.storage_widget.get()

    def to_object(self):
        data = self.get_gui_data()  # List items are not to be converted
        for d in data:
            if isinstance(d, str) and str not in self._list_args:
                self.check_literals(d, self.filter_literals(self._list_args))

        return get_origin(self.class_)(data)

    def _edit_selected(self):
        selection = self.storage_widget.curselection()
        if len(selection) == 1:
            object_ = self.storage_widget.get()[selection[0]]
            if isinstance(object_, ObjectInfo):
                self.new_object_frame(object_.class_, self.storage_widget, old_data=object_)
            else:
                self.new_object_frame(type(object_), self.storage_widget, old_data=object_)
        else:
            Messagebox.show_error("Selection error", "Select ONE item!", parent=self)


class NewObjectFrameIterableView(NewObjectFrameIterable):
    """
    Same as :class:`NewObjectFrameIterable`, but only allows viewing iterable objects.
    Also does not require annotations as supposed to the base class.
    """
    def __init__(self, class_: Any, return_widget: Any, parent=None, old_data: list = None, check_parameters: bool = True, allow_save=True):
        dpi_5 = dpi_scaled(5)
        super(NewObjectFrameBase, self).__init__(class_, return_widget, parent, old_data, check_parameters, allow_save)
        self.storage_widget = w = ListBoxScrolled(self.frame_main, height=20)
        ListboxTooltip(self.storage_widget, 0)

        frame_edit_remove = ttk.Frame(self.frame_main, padding=(dpi_5, 0))
        frame_edit_remove.pack(side="right")
        frame_cp = ttk.Frame(frame_edit_remove)
        frame_cp.pack(fill=tk.X, expand=True, pady=dpi_5)

        ttk.Button(frame_edit_remove, text="View", command=lambda: self._edit_selected()).pack(fill=tk.X)
        w.pack(side="left", fill=tk.BOTH, expand=True)

        if old_data is not None:
            self.load(old_data)

        self.remember_gui_data()



"""
Module contains functionality for managing type aliases.
Type alias is an alternative type name.

E. g., we could have an alias for timedelta -> Duration.
"""
from typing import Dict, Optional
from .doc import doc_category


__all__ = (
    "register_alias",
    "get_aliased_name",
)


ALIASES: Dict[type, str] = {}


@doc_category("Aliasing")
def register_alias(cls: type, alias: str):
    """
    Creates an alias name for type ``cls``.
    Aliases will then be displayed in place of original name in:

    - Object definition "New" menu
    - Container values (e. g., inside Combobox)

    Parameters
    ------------
    cls: type
        The class (or function or type) to register an alias for.
    alias: str
        The alias.

    Example
    -----------
    
    .. code-block:: python

        from datetime import timedelta

        register_alias(timedelta, "Duration")
    """
    ALIASES[cls] = alias


@doc_category("Aliasing")
def get_aliased_name(class_) -> Optional[str]:
    """
    Returns the ``class_``'s aliased name. If class has no alias name,
    ``None`` is returned.
    """
    return ALIASES.get(class_)



"""
Module used for managing annotations.
"""
from typing import Union, Optional, Generic, Iterable, get_args, get_origin, get_type_hints
from datetime import datetime, timedelta, timezone
from inspect import isclass, isabstract
from itertools import product, chain
from contextlib import suppress
try:
    from types import UnionType
except ImportError:
    UnionType = Union

from .utilities import issubclass_noexcept
from .doc import doc_category


__all__ = (
    "register_annotations",
    "get_annotations",
    "convert_types",
)


ADDITIONAL_ANNOTATIONS = {
    timedelta: {
        "days": float,
        "seconds": float,
        "microseconds": float,
        "milliseconds": float,
        "minutes": float,
        "hours": float,
        "weeks": float
    },
    datetime: {
        "year": int,
        "month": Union[int, None],
        "day": Union[int, None],
        "hour": int,
        "minute": int,
        "second": int,
        "microsecond": int,
        "tzinfo": timezone
    },
    timezone: {
        "offset": timedelta,
        "name": str
    },
}


@doc_category("Annotations")
def register_annotations(cls: type, mapping: Optional[dict] = {}, **annotations):
    """
    Extends original annotations of ``cls``.

    This can be useful eg. when the class your
    are adding is not part of your code and is also not annotated.

    Classes that already have additional annotations:

    - datetime.timedelta
    - datetime.datetime 
    - datetime.timezone

    Parameters
    ------------
    cls: type
        The class (or function) to register additional annotations on.
    mapping: Optional[Dict[str, type]]
        Mapping mapping the parameter name to it's type.
    annotations: Optional[Unpack[str, type]]
        Keyword arguments mapping parameter name to it's type (``name=type``).

    Example
    -----------
    
    .. code-block:: python

        from datetime import timedelta, timezone

        register_annotations(
            timezone,
            offset=timedelta,
            name=str
        )    
    """
    if cls not in ADDITIONAL_ANNOTATIONS:
        ADDITIONAL_ANNOTATIONS[cls] = {}

    ADDITIONAL_ANNOTATIONS[cls].update(**annotations, **mapping)


@doc_category("Annotations")
def get_annotations(class_) -> dict:
    """
    Returns class / function annotations including the ones extended with ``register_annotations``.
    It does not return the return annotation.

    Additionally, this function resolves any generic types to their parameterized types, but
    only for classes, functions don't support this yet as support for generics on functions was added
    in Python 3.12.
    """
    annotations = {}
    with suppress(AttributeError, TypeError):
        if isclass(class_):
            annotations = get_type_hints(class_.__init__)
        elif isclass(origin_class := get_origin(class_)) and issubclass_noexcept(origin_class, Generic):
            # Resolve generics
            annotations = get_type_hints(origin_class.__init__)
            generic_types = get_args(origin_class.__orig_bases__[0])
            generic_values = get_args(class_)
            generic_name_value = {generic_types[i]: generic_values[i] for i in range(len(generic_types))}
            for k, v in annotations.items():
                annotations[k] = generic_name_value.get(v, v)
        else:
            annotations = get_type_hints(class_)

    additional_annotations = ADDITIONAL_ANNOTATIONS.get(class_, {})
    annotations = {**annotations, **additional_annotations}

    if "return" in annotations:
        del annotations["return"]

    return annotations


def convert_types(input_type: type):
    """
    Type preprocessing method, that extends the list of types with inherited members (polymorphism)
    and removes classes that are wrapped by some other class, if the wrapper class also appears in
    the annotations.
    """
    def remove_classes(types: list):
        r = types.copy()
        for type_ in types:
            # It's a wrapper of some class -> remove the wrapped class
            if hasattr(type_, "__wrapped__"):
                if type_.__wrapped__ in r:
                    r.remove(type_.__wrapped__)

            # Abstract classes are classes that don't allow instantiation -> remove the class
            if isabstract(type_):
                r.remove(type_)

        return tuple({a:0 for a in r})


    if isinstance(input_type, str):
        raise TypeError(
            f"Provided type '{input_type}' is not a type - it is a string!\n"
            "Potential subscripted type problem?\n"
            "Instead of e. g., list['type'], try using typing.List['type']."
        )

    origin = get_origin(input_type)
    if issubclass_noexcept(origin, Generic):
        # Patch for Python versions < 3.10
        input_type.__name__ = origin.__name__

    # Unpack Union items into a tuple
    union_type = origin in {Union, UnionType}
    if union_type or issubclass_noexcept(origin, (Iterable, Generic)):
        new_types = []
        for arg_group in get_args(input_type):
            new_types.append(remove_classes(list(convert_types(arg_group))))

        if union_type:
            return tuple(chain.from_iterable(new_types))  # Just expand unions

        # Process abstract classes and polymorphism
        new_origins = []
        for origin in convert_types(origin):
            if issubclass_noexcept(origin, Generic):
                for comb in product(*new_types):
                    new_origins.append(origin[comb])
            elif issubclass_noexcept(origin, Iterable):
                new = origin[tuple(chain.from_iterable(new_types))] if len(new_types) else origin
                new_origins.append(new)
            else:
                new_origins.append(origin)

        return remove_classes(new_origins)

    if input_type.__module__ == "builtins":
        # Don't consider built-int types for polymorphism
        # No removal of abstract classes is needed either as builtins types aren't abstract
        return (input_type,)

    # Extend subclasses
    subtypes = []
    if hasattr(input_type, "__subclasses__"):
        for st in input_type.__subclasses__():
            subtypes.extend(convert_types(st))

    # Remove wrapped classes (eg. wrapped by decorator) + ABC classes
    return remove_classes([input_type, *subtypes])



from typing import get_args, get_origin, Iterable, Union, Literal, Dict, Tuple
from functools import partial
from enum import Enum, Flag


from ..convert import *
from ..dpi import *
from ..utilities import *
from ..storage import *
from ..messagebox import Messagebox
from ..extensions import extendable
from ..annotations import get_annotations, convert_types
from ..deprecation import *
from ..doc import doc_category

from .frame_base import *
from .tooltip import ComboboxTooltip

import tkinter as tk
import tkinter.ttk as ttk
import tkinter.filedialog as tkfile

import inspect
import copy
import json


__all__ = (
    "NewObjectFrameStruct",
    "NewObjectFrameStructView",
)


@extendable
@doc_category("Object frames")
class NewObjectFrameStruct(NewObjectFrameBase):
    """
    Frame for inside the :class:`ObjectEditWindow` that allows object definition.

    Parameters
    -------------
    class_: Any
        The class we are defining for.
    return_widget: ComboBoxObjects | ListBoxScrolled
        The widget to insert the ObjectInfo into after saving.
    parent: TopLevel
        The parent window.
    old_data: ObjectInfo
        The old_data ObjectInfo object to edit.
    check_parameters: bool
        Check parameters (by creating the real object) upon saving.
        This is ignored if editing a function instead of a class.
    allow_save: bool
        If False, will open in read-only mode.
    additional_values: Dict[str, Any]
        A mapping of additional values to be inserted into corresponding field.
    """
    def __new__(cls, *args, **kwargs):
        if kwargs.get("allow_save", True):
            obj = super().__new__(NewObjectFrameStruct)
        else:
            obj = super().__new__(NewObjectFrameStructView)

        return obj

    def __init__(
        self,
        class_,
        return_widget: Union[ComboBoxObjects, ListBoxScrolled],
        parent = None,
        old_data: ObjectInfo = None,
        check_parameters: bool = True,
        allow_save = True,
        additional_values: dict = {},
        _annotations_override: dict = None,
    ):
        super().__init__(class_, return_widget, parent, old_data, check_parameters,allow_save)
        self._map: Dict[str, Tuple[ComboBoxObjects, Iterable[type]]] = {}
        dpi_5 = dpi_scaled(5)
        dpi_5_h = dpi_5 // 2

        if not (annotations := _annotations_override or get_annotations(class_)):
            raise TypeError("This object cannot be edited.")

        # Template
        @gui_except(window=self)
        def save_template():
            filename = tkfile.asksaveasfilename(filetypes=[("JSON", "*.json")], parent=self)
            if filename == "":
                return

            json_data = convert_to_dict(self.to_object(ignore_checks=True))

            if not filename.endswith(".json"):
                filename += ".json"

            with open(filename, "w", encoding="utf-8") as file:
                json.dump(json_data, file, indent=2)

            Messagebox.show_info("Finished", f"Saved to {filename}", parent=self)

        @gui_except(window=self)
        def load_template():
            filename = tkfile.askopenfilename(filetypes=[("JSON", "*.json")], parent=self)
            if filename == "":
                return

            with open(filename, "r", encoding="utf-8") as file:
                json_data: dict = json.loads(file.read())
                object_info = convert_from_dict(json_data)
                # Get class_ attribute if we have the ObjectInfo type, if not just compare the actual type
                if object_info.class_ is not self.class_:
                    raise TypeError(
                        f"The selected template is not a {self.class_.__name__} template.\n"
                        f"The requested template is for type: {object_info.class_.__name__}!"
                    )

                self.load(object_info)

        bnt_menu_template = ttk.Menubutton(self.frame_toolbar, text="Template")
        menu = tk.Menu(bnt_menu_template)
        menu.add_command(label="Load template", command=load_template)
        menu.add_command(label="Save template", command=save_template)
        bnt_menu_template.configure(menu=menu)
        bnt_menu_template.pack(side="left")

        # Nickname entry
        self.entry_nick = HintedEntry(
            "Object nickname",
            self.frame_main,
            state="normal" if self.allow_save else "disabled"
        )
        self.entry_nick.pack(anchor=tk.W, padx=dpi_5_h, pady=dpi_5)
        annotations_depr = {k:v for k,v in annotations.items() if is_deprecated(class_, k)}
        for k in annotations_depr:
            del annotations[k]

        self._create_fields(annotations, additional_values, self.frame_main)
        if annotations_depr:
            ttk.Separator(self.frame_main).pack(fill=tk.X, pady=dpi_5)
            ttk.Label(self.frame_main, text="Deprecated", font=("TkDefaultFont", 10)).pack(anchor=tk.W)
            self._create_fields(annotations_depr, additional_values, self.frame_main)

        if old_data is not None:  # Edit
            self.load(old_data)

        self.remember_gui_data()

    def _create_fields(self, annotations: dict[str, type], additional_values: dict, frame: ttk.Frame):
        dpi_5 = dpi_scaled(5)
        dpi_5h = dpi_5 // 2

        labels: list[ttk.Label] = []
        for (k, v) in annotations.items():
            # Init widgets
            entry_types = convert_types(v)
            frame_annotated = ttk.Frame(frame)
            frame_annotated.pack(fill=tk.BOTH, expand=True, pady=dpi_5)
            label = ttk.Label(frame_annotated, text=k)
            labels.append(label)
            label.pack(side="left")

            # Storage widget with the tooltip for displaying
            # nicknames on ObjectInfo instances
            w = combo = ComboBoxObjects(frame_annotated)
            ComboboxTooltip(w)

            bnt_new_menu = ttk.Menubutton(frame_annotated, text="New")
            menu_new = tk.Menu(bnt_new_menu)
            bnt_new_menu.configure(menu=menu_new)

            deprecated_param_types = set(t for t in entry_types if is_deprecated(self.class_, k, t))
            normal_types = [t for t in entry_types if t not in deprecated_param_types]
            any_filled = self._fill_field_values(k, normal_types, menu_new, combo, additional_values)

            if deprecated_param_types:
                menu_depr = tk.Menu(menu_new)
                menu_new.add_cascade(label=">Deprecated", menu=menu_depr)
                any_filled = self._fill_field_values(
                    k,
                    list(deprecated_param_types),
                    menu_depr,
                    combo,
                    additional_values
                ) or any_filled

            bnt_edit = ttk.Button(
                frame_annotated,
                text="🖋️",
                width=3,
                command=partial(self._edit_selected, k, w)
            )

            bnt_copy_paste = ttk.Menubutton(frame_annotated, text="C/P")
            copy_menu = tk.Menu(bnt_copy_paste)
            copy_menu.add_command(label="Copy", command=combo.save_to_clipboard)
            copy_menu.add_command(label="Paste", command=combo.paste_from_clipboard)
            bnt_copy_paste.configure(menu=copy_menu)

            if not (any_filled and self.allow_save):
                bnt_new_menu.configure(state="disabled")

            bnt_copy_paste.pack(side="right", padx=dpi_5h)
            bnt_edit.pack(side="right", padx=dpi_5h)
            bnt_new_menu.pack(side="right", padx=dpi_5h)
            combo.pack(fill=tk.X, side="right", expand=True, padx=dpi_5h)
            self._map[k] = (w, entry_types)

        max_width = max(label.winfo_reqwidth() for label in labels) // 5
        for label in labels:
            label.config(width=max_width)

    def _fill_field_values(
        self,
        k: str,
        entry_types: list,
        menu: tk.Menu,
        combo: ComboBoxObjects,
        additional_values: dict
    ):
        "Fill ComboBox values based on types in ``entry_types`` and create New <object_type> buttons"
        any_filled = False
        for entry_type in entry_types:
            if get_origin(entry_type) is Literal:
                values = get_args(entry_type)
                combo["values"] = values
                # tkvalid.add_option_validation(combo, values)
            elif entry_type is bool:
                combo.insert(tk.END, True)
                combo.insert(tk.END, False)
                # tkvalid.add_option_validation(combo, ["True", "False", ''])
            elif issubclass_noexcept(entry_type, Enum) and not issubclass_noexcept(entry_type, Flag):
                combo["values"] = values = [en for en in entry_type]
                # tkvalid.add_option_validation(combo, list(map(str, values)) + [''])
            elif entry_type is type(None):
                if bool not in entry_types:
                    combo.insert(tk.END, None)
            else:  # Type not supported, try other types
                any_filled = True
                if self.allow_save:
                    menu.add_command(
                        label=f"New {self.get_cls_name(entry_type, args=True)}",
                        command=partial(self.new_object_frame, entry_type, combo)
                    )

        # Additional values to be inserted into ComboBox
        for value in additional_values.get(k, []):
            combo.insert(tk.END, value)

        # The class of last list like type. Needed when "Edit selected" is used
        # since we don't know what type it was
        return any_filled

    @extendable
    def load(self, old_data: ObjectInfo):
        data = old_data.data
        for attr, (widget, types_) in self._map.items():
            if attr not in data:
                continue

            val = data[attr]

            if val not in widget["values"]:
                widget.insert(tk.END, val)

            widget.current(widget["values"].index(val))

        if old_data.nickname:
            self.entry_nick.insert('0', old_data.nickname)

        self.old_gui_data = old_data

    @extendable
    def to_object(self, *, ignore_checks = False) -> ObjectInfo:
        """
        Converts GUI data into an ObjectInfo abstraction object.

        Parameters
        ------------
        ignore_checks: bool
            Don't check the correctness of given parameters.
        """
        map_ = {}
        widget: ComboBoxObjects
        for attr, (widget, types_) in self._map.items():
            value = widget.get()
            # Either it's a string needing conversion or a Literal constant not to be converted
            if isinstance(value, str):
                if not value:
                    continue

                try:
                    value = self.cast_type(value, types_)
                except TypeError as exc:
                    # Perhaps it's a valid literal:
                    literal_types = self.filter_literals(types_)
                    if not literal_types:
                        raise

                    self.check_literals(value, literal_types)

            map_[attr] = value

        class_ = get_origin(self.class_) or self.class_
        nickname = self.entry_nick.get() or None
        object_ = ObjectInfo(class_, map_, nickname)  # Abstraction of the underlying object
        if (
            not ignore_checks and
            self.check_parameters and
            (inspect.isclass(class_))  # Only check objects
        ):
            # Cache the object created for faster
            _convert_to_objects_cached(object_)  # Tries to create instances to check for errors

        return object_

    def get_gui_data(self) -> dict:
        values = {}
        for attr, (widget, types_) in self._map.items():
            value = widget.get()
            if isinstance(value, str) and not value:  # Ignore empty values
                continue

            if isinstance(value, (list, tuple, set)):  # Copy lists else the values would change in the original state
                value = copy.copy(value)

            values[attr] = value

        return values

    @gui_except()
    def _edit_selected(self, key: str, combo: ComboBoxObjects):
        selection = combo.get()

        # Convert selection to any of the allowed types.
        # This is used for casting manually typed values (which are strings) into appropriate types
        # Eg. if a number is manually typed in and Edit button is clicked, this would result in a string type
        # edit request, which is not really the intend. The intend is to edit a number in this example.
        if isinstance(selection, str):
            selection = self.cast_type(selection, self._map[key][1])

        if isinstance(selection, ObjectInfo):
            return self.new_object_frame(selection.class_, combo, old_data=selection)
        else:
            type_sel = type(selection)
            for t in convert_types(get_annotations(self.class_)[key]):
                if (get_origin(t) or t) == type_sel:
                    type_ = t
                    break
            else:
                type_ = None

            return self.new_object_frame(type_, combo, old_data=selection)


class NewObjectFrameStructView(NewObjectFrameStruct):
    """
    Same as :class:`NewObjectFrameStruct`, but creates a wrapper Viewer class, which's attributes automatically
    get mapped as annotations, which the :class:`NewObjectFrameStruct` knows how to handle.
    """
    def __init__(self, class_, *args, **kwargs):
        old_data: ObjectInfo = kwargs["old_data"]
        annotations = {k: v.class_ if isinstance(v, ObjectInfo) else type(v) for k, v in old_data.data.items()}
        super().__init__(class_, *args, **kwargs, _annotations_override=annotations)

    @gui_except()
    def _edit_selected(self, key: str, combo: ComboBoxObjects):
        selection = combo.get()

        # Convert selection to any of the allowed types.
        # This is used for casting manually typed values (which are strings) into appropriate types
        # Eg. if a number is manually typed in and Edit button is clicked, this would result in a string type
        # edit request, which is not really the intend. The intend is to edit a number in this example.
        if isinstance(selection, str):
            selection = self.cast_type(selection, self._map[key][1])

        if isinstance(selection, ObjectInfo):
            return self.new_object_frame(selection.class_, combo, old_data=selection)
        else:
            return self.new_object_frame(type(selection), combo, old_data=selection)



"""
Utility module.
"""
from typing import Any, Callable, Tuple
from functools import wraps

import importlib
from .messagebox import Messagebox
from .doc import doc_category, DOCUMENTATION_MODE


__all__ = (
    "import_class",
    "gui_except",
    "gui_confirm_action",
    "issubclass_noexcept",
)


@doc_category("Utilities")
def import_class(path: str) -> type:
    """
    Imports the class provided by it's ``path``.
    """
    path = path.split(".")
    module_path, class_name = '.'.join(path[:-1]), path[-1]
    try:
        module = importlib.import_module(module_path)
    except Exception as exc:  # Fall back for older versions, try to import from package instead of module
        module_path = module_path.split('.')[0]
        module = importlib.import_module(module_path)

    class_ = getattr(module, class_name)
    return class_


@doc_category("Utilities")
def gui_except(window = None):
    """
    Function that returns a decorator.
    The decorator returns a callable wrapper object.

    When the wrapper is called, it calls the original callee, and in case an exception is raised,
    the exception is displayed through a popup window.
    If no exception is raised, the callee's result is returned.


    Parameters
    --------------
    window: Optional[Any]
        Optional parent widget, if not provided, the owner of callee method will be used.

    Example
    ----------

    .. code-block:: python
    
        class MyWindow(tk.TopLevel):
            
            @gui_except()
            def raise_exceptions(self):
                raise ValueError("Test")
    """
    def decorator(fnc: Callable):
        if DOCUMENTATION_MODE:
            return fnc

        @wraps(fnc, updated=[])
        class wrapper:
            def __init__(self, instance = None) -> None:
                self.inst = instance
            
            def __get__(self, instance, cls = None):
                return wrapper(instance)
            
            def __call__(self, *args: Any, **kwargs: Any) -> Any:
                parent = window or self.inst
                try:
                    if self.inst is not None:
                        args = (self.inst, *args)

                    return fnc(*args, **kwargs)
                except Exception as exc:
                    Messagebox.show_error(f"Exception in {fnc.__name__}", str(exc), parent=parent)

        return wrapper()

    return decorator


@doc_category("Utilities")
def gui_confirm_action(parent = None):
    """
    Function that returns a decorator.
    The decorator returns a callable wrapper object.

    When the wrapper is called, the user is asked to confirm the action and if the user confirms,
    the original function is called and the wrapper returns the result.
    If action is cancelled, None is returned by the wrapper.

    Parameters
    --------------
    parent: Optional[Any]
        Optional parent widget, if not provided, the owner of callee method will be used.


    Example
    -----------

    .. code-block:: python

        class Database(tk.Widget):
            @gui_confirm_action()
            def delete_database(self):
                print("Deleting database")
    """
    def _gui_confirm_action(fnc: Callable):
        if DOCUMENTATION_MODE:
            return fnc

        class wrapper:
            def __init__(self, bind = None) -> None:
                self.bind = bind

            def __call__(self, *args, **kwargs):
                result = Messagebox.yesnocancel("Confirm", "Are you sure?", parent=parent or self.bind)

                if self.bind is not None:
                    args = (self.bind, *args)  # self, *args

                if result:
                    return fnc(*args, **kwargs)
                
            def __get__(self, instance, cls = None):
                return wrapper(instance)

        return wrapper()

    return _gui_confirm_action


@doc_category("Utilities")
def issubclass_noexcept(cls: type, bases: Tuple[type]):
    """
    Same as :func:`inspect.issubclass`, but instead of raising an exception
    when the second argument is not a class, it returns False.
    """
    try:
        return issubclass(cls, bases)
    except Exception:
        return False



from typing import get_origin
from inspect import isclass, isfunction
from collections.abc import Sequence, Set
from enum import Flag

from .frame_number import *
from .frame_iterable import *
from .frame_string import *
from .frame_struct import *
from .frame_base import *
from .frame_flag import *

from ..utilities import gui_except, issubclass_noexcept
from ..extensions import extendable
from ..doc import doc_category
from ..dpi import dpi_scaled

import tkinter.ttk as ttk
import tkinter as tk


__all__ = (
    "ObjectEditWindow",
)


@extendable
@doc_category("Object window")
class ObjectEditWindow(tk.Toplevel):
    """
    Top level window for creating and editing new objects.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._closed = False

        dpi_5 = dpi_scaled(5)

        # Elements
        self.opened_frames: list[NewObjectFrameBase] = []
        self.frame_main = ttk.Frame(self, padding=(dpi_5, dpi_5))
        self.frame_toolbar = ttk.Frame(self, padding=(dpi_5, dpi_5))
        ttk.Button(self.frame_toolbar, text="Close", command=self.close_object_edit_frame).pack(side="left")
        ttk.Button(self.frame_toolbar, text="Save", command=self.save_object_edit_frame).pack(side="left")

        self.frame_toolbar.pack(expand=False, fill=tk.X)
        self.frame_main.pack(expand=True, fill=tk.BOTH)
        self.frame_main.rowconfigure(0, weight=1)
        self.frame_main.columnconfigure(0, weight=1)

        var = tk.BooleanVar(value=True)
        ttk.Checkbutton(
            self.frame_toolbar,
            text="Keep on top",
            variable=var,
            command=lambda: self.attributes("-topmost", var.get()),
        ).pack(side="right")
        self.attributes("-topmost", var.get())

        # Window initialization
        NewObjectFrameBase.set_origin_window(self)
        self.protocol("WM_DELETE_WINDOW", self.close_object_edit_frame)

    @property
    def closed(self) -> bool:
        return self._closed

    def open_object_edit_frame(
        self,
        class_,
        return_widget,
        old_data = None,
        check_parameters: bool = True,
        allow_save: bool = True,
        **kwargs
    ):
        """
        Opens new frame for defining an object.
        Parameters are the same as for NewObjectFrameBase.

        Parameters
        ------------
        class_: Any
            The class we are defining for.
        return_widget: Any
            The widget to insert the ObjectInfo into after saving.
        old_data: Any
            The old_data gui data.
        check_parameters: bool
            Check parameters (by creating the real object) upon saving.
            This is ignored if editing a function instead of a class.
        allow_save: bool
            If False, will open in read-only mode.
        """
        frame = self._create_and_add_frame(class_, return_widget, old_data, check_parameters, allow_save, kwargs)
        if frame is None and not self.opened_frames:
            self.destroy()
            self._closed = True

    @gui_except()
    def _create_and_add_frame(
        self,
        class_: type,
        return_widget,
        old_data,
        check_parameters: bool,
        allow_save: bool,
        kwargs
    ):
        frame: NewObjectFrameBase
        class_origin = get_origin(class_) or class_  # Remove any Generic type subscriptions

        if issubclass_noexcept(class_origin, Flag):
            frame_class = NewObjectFrameFlag
        elif issubclass_noexcept(class_origin, str):
            frame_class = NewObjectFrameString
        elif issubclass_noexcept(class_origin, (int, float, complex)):
            frame_class = NewObjectFrameNumber
        elif issubclass_noexcept(class_origin, (Sequence, Set)):
            frame_class = NewObjectFrameIterable
        else:
            frame_class = NewObjectFrameStruct

        self.opened_frames.append(
            frame := frame_class(
                class_,
                return_widget,
                old_data=old_data,
                check_parameters=check_parameters,
                allow_save=allow_save,
                parent=self.frame_main,
                **kwargs
            )
        )

        if len(self.opened_frames) > 1:
            self.opened_frames[-2].pack_forget()

        frame.pack(fill=tk.BOTH, expand=True)
        frame.update_window_title()
        self.set_default_size_y()

        return frame

    def close_object_edit_frame(self):
        self.opened_frames[-1].close_frame()

    def save_object_edit_frame(self):
        self.opened_frames[-1].save()

    def clean_object_edit_frame(self):
        self.opened_frames.pop().destroy()
        opened_frames_len = len(self.opened_frames)

        if opened_frames_len:
            frame = self.opened_frames[-1]
            frame.pack(fill=tk.BOTH, expand=True)  # (row=0, column=0)
            frame.update_window_title()
            self.set_default_size_y()
        else:
            self._closed = True
            self.destroy()

    def set_default_size_y(self):
        "Sets window Y size to default"
        self.update()
        self.geometry(f"{self.winfo_width()}x{self.winfo_reqheight()}")



"""
Module contains definitions for automatic scaling over different DPI displays.
"""
DPI_ORIGINAL = 96


class GLOBALS:
    current_dpi = 96


__all__ = (
    "dpi_scaled",
    "set_dpi"
)


def dpi_scaled(px: int) -> int:
    return int(px * GLOBALS.current_dpi / DPI_ORIGINAL)


def set_dpi(new_dpi: int):
    GLOBALS.current_dpi = new_dpi




from typing import Any

from ..storage import *
from .frame_base import *
from ..extensions import extendable
from ..doc import doc_category

import tkinter as tk


TEXT_MAX_UNDO = 20


__all__ = (
    "NewObjectFrameString",
)


@extendable
@doc_category("Object frames")
class NewObjectFrameString(NewObjectFrameBase):
    def __init__(
        self,
        class_: Any,
        return_widget: Any,
        parent=None,
        old_data: str = None,
        check_parameters: bool = True,
        allow_save=True
    ):
        super().__init__(class_, return_widget, parent, old_data, check_parameters, allow_save)
        self.storage_widget = Text(self.frame_main, undo=True, maxundo=TEXT_MAX_UNDO)
        self.storage_widget.pack(fill=tk.BOTH, expand=True)

        if old_data is not None:
            self.load(old_data)

        self.remember_gui_data()

    def load(self, old_data: Any):
        self.storage_widget.insert(tk.END, old_data)
        self.old_gui_data = old_data

    def get_gui_data(self) -> Any:
        return self.storage_widget.get()

    def to_object(self):
        return self.get_gui_data()




"""
Wrapper module, used to add support
for both tkinter and ttkbootstrap message boxes.

This module needs to be imported after ttkbootstrap!
"""
import sys
if sys.modules.get('ttkbootstrap') is not None:
    from ttkbootstrap.dialogs.dialogs import Messagebox as BootMb
    TTKBOOT_INSTALLED = True
else:
    import tkinter.messagebox as mb
    TTKBOOT_INSTALLED = False


__all__ = ("Messagebox",)


class Messagebox:
    """
    Wrapper for some of Messagebox methods, that offers compatibility between
    ttk and ttkbootstrap.
    """
    def _process_kwargs(kwargs):
        if "master" in kwargs:
            kwargs['parent'] = kwargs["master"]
            del kwargs["master"]

    if TTKBOOT_INSTALLED:
        @classmethod
        def yesnocancel(cls, title: str, message: str, **kwargs):
            cls._process_kwargs(kwargs)
            r = BootMb.yesnocancel(message, title, **kwargs)
            if r is not None and r != 'Cancel':
                return r == 'Yes'
        
        @classmethod
        def show_error(cls, title: str, message: str, **kwargs):
            cls._process_kwargs(kwargs)
            BootMb.show_error(message, title, **kwargs)

        @classmethod
        def show_info(cls, title: str, message: str, **kwargs):
            cls._process_kwargs(kwargs)
            BootMb.show_info(message, title, **kwargs)
    else:
        @classmethod
        def yesnocancel(cls, title: str, message: str, **kwargs):
            cls._process_kwargs(kwargs)
            return mb.askyesnocancel(title, message, **kwargs)

        @classmethod
        def show_error(cls, title: str, message: str, **kwargs):
            cls._process_kwargs(kwargs)
            mb.showerror(title, message, **kwargs)

        @classmethod
        def show_info(cls, title: str, message: str, **kwargs):
            cls._process_kwargs(kwargs)
            mb.showinfo(title, message, **kwargs)


"""
Modules contains storage container widgets.
"""
from typing import Union, Any, List, Iterable, Callable

from .utilities import gui_confirm_action
from .messagebox import Messagebox
from .convert import ObjectInfo
from .doc import doc_category

import tkinter.ttk as ttk
import tkinter as tk


__all__ = (
    "Text",
    "ListBoxObjects",
    "ListBoxScrolled",
    "ComboBoxObjects",
    "ComboEditFrame",
    "HintedEntry",
    "PyObjectScalar",
)


class _NoClipBoard:
    pass


class GLOBAL:
    clipboard = _NoClipBoard


@doc_category("Storage widgets")
class HintedEntry(ttk.Entry):
    """
    A hinted tkinter.ttk.Entry.
    The entry accepts the parameter ``hint`` (The hinting text which will be displayed in gray)
    and the original tkinter.ttk.Entry parameters.
    """
    def __init__(self, hint: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hint = hint
        self._set_hint()
        self.bind("<FocusIn>", self._focus_in)
        self.bind("<FocusOut>", self._focus_out)

    def get(self) -> str:
        if str(self["foreground"]) == "gray":
            return ''

        return super().get()

    def insert(self, *args, **kwargs) -> None:
        state = self.cget("state")
        self.config(state="enabled")
        if str(self["foreground"]) == "gray":
            self.delete('0', tk.END)

        _ret = super().insert(*args, **kwargs)
        self["foreground"] = "black"
        self.config(state=state)
        return _ret

    def _set_hint(self):
        self.insert('0', self.hint)
        self["foreground"] = "gray"

    def _focus_in(self, event: tk.Event):
        if str(self["foreground"]) == "gray":
            self["foreground"] = 'black'
            self.delete('0', tk.END)

    def _focus_out(self, event: tk.Event):
        if not super().get():
            self._set_hint()


@doc_category("Storage widgets")
class PyObjectScalar(ttk.Frame):
    """
    Represents a single storage widget for a Python object.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.element = None
        self.display = ttk.Entry(self, state="readonly")
        self.display.pack(fill=tk.BOTH, expand=True)

    def get(self) -> object:
        """
        Returns the stored Python object (or None if not set).
        """
        return self.element

    def set(self, value: object):
        """
        Sets the Python object as the widget's value.
        """
        self.display.config(state="active")
        self.display.delete("0", tk.END)
        self.display.insert(tk.END, str(value))
        self.display.config(state="readonly")
        self.element = value


@doc_category("Storage widgets")
class Text(tk.Text):
    """
    Modified version of :class`tkinter.Text`, which replaces the original
    ``get`` method to contain no parameters and instead return the entire text stripped.
    This is for compatibility with other storage widgets.
    """
    def get(self) -> str:
        "Returns entire text stripped."
        return super().get("1.0", tk.END).strip()


@doc_category("Storage widgets")
class ListBoxObjects(tk.Listbox):
    """
    Modified version of :class:`tkinter.Listbox`, which:

    - Allows copy-paste and also sets the standard keyboard shortcuts for it
    - Implements the ``current`` method for compatibility with :class:`tkclasswiz.storage.ComboBoxObjects`
    - Overwrites the ``get`` method to return the original objects instead of text
    - Modifies the ``insert`` method to insert original objects instead of text
    """
    def __init__(self, *args, **kwargs):
        self._original_items = []
        super().__init__(*args, **kwargs)
        self.configure(selectmode=tk.EXTENDED)

        self.bind("<Control-c>", lambda e: self.save_to_clipboard())
        self.bind("<BackSpace>", lambda e: self.delete_selected())
        self.bind("<Delete>", lambda e: self.delete_selected())
        self.bind("<Control-v>", lambda e: self.paste_from_clipboard())

    def current(self) -> int:
        "Returns index of the first currently selected element or -1 if none selected."
        selection = self.curselection()
        return selection[0] if len(selection) else -1

    def get(self, first: int = 0, last: int = None) -> list:
        """
        Returns elements from ``first`` to ``last`` (not included).
        """
        slice_range = slice(first, last)
        return self._original_items[slice_range]

    def insert(self, index: Union[str, int], *elements: Union[str, float]) -> None:
        """
        Inserts ``elements`` to ``index``. The ``index`` must be a numerical index or tkinter.END ('end').
        """
        _ret = super().insert(index, *elements)
        if isinstance(index, str):
            index = len(self._original_items) if index == "end" else 0

        old_data = self._original_items[index:]
        self._original_items[index:] = list(elements) + old_data

        return _ret

    def delete(self, *indexes: int) -> None:
        """
        Delete elements that have the index in ``indexes``
        """
        def create_ranges() -> List[tuple]:
            start_i = indexes[0]
            len_indexes = len(indexes)
            for i in range(1, len_indexes):
                if indexes[i] > indexes[i - 1] + 1:  # Current index more than 1 bigger than previous
                    to_yield = start_i, indexes[i - 1]
                    yield to_yield

                    # After element is erased, all items' indexes get shifted to the left
                    # from the deleted item forward (right) -> Need to shift our list of indexes also.
                    for j in range(i, len_indexes):
                        indexes[j] -= (to_yield[1] - to_yield[0] + 1)

                    start_i = indexes[i]

            yield start_i, indexes[-1]

        if indexes[-1] == "end":
            indexes = range(indexes[0], len(self._original_items))

        indexes = sorted(list(indexes))
        for range_ in create_ranges():
            super().delete(*range_)
            del self._original_items[range_[0]:range_[1] + 1]

    def count(self) -> int:
        """
        Returns the amount of elements inside the internal list.
        """
        return len(self._original_items)

    @gui_confirm_action()
    def delete_selected(self):
        """
        Deletes selected elements.
        """
        sel: List[int] = self.curselection()
        if len(sel):
            self.delete(*sel)
        else:
            Messagebox.show_error("Empty list!", "Select atleast one item!", parent=self)

    def clear(self) -> None:
        """
        Deletes all the elements inside the internal list.
        """
        super().delete(0, tk.END)
        self._original_items.clear()

    def save_to_clipboard(self):
        """
        Saves selection to clipboard.
        """
        selection = self.curselection()
        if len(selection):
            object_: Union[ObjectInfo, Any] = self.get()[min(selection):max(selection) + 1]
            GLOBAL.clipboard = object_ if len(selection) > 1 else object_[0]
        else:
            Messagebox.show_error("Empty list!", "Select atleast one item!", parent=self)

    def paste_from_clipboard(self):
        """
        Paste elements from clipboard.
        """
        if GLOBAL.clipboard is _NoClipBoard:
            return  # Clipboard empty

        if isinstance(GLOBAL.clipboard, Iterable):
            self.insert(tk.END, *GLOBAL.clipboard)
        else:
            self.insert(tk.END, GLOBAL.clipboard)

    def move(self, index: int, direction: int):
        """
        Move a element inside the list box around.

        Parameters
        --------------
        index: int
            Index of the element to move.
        direction: int
            To move forward pass 1, to move backwards pass -1.
        """
        if direction == -1 and index == 0 or direction == 1 and index == len(self._original_items) - 1:
            return

        value = self._original_items[index]
        self.delete(index)
        index += direction
        self.insert(index, value)
        self.selection_set(index)

    def move_selection(self, direction: int):
        """
        Moves the selected element up inside the list box.
        Pass ``direction`` value 1 if you want to move forward, or -1 for backwards.
        """
        if len(selection := self.curselection()) == 1:
            self.move(selection[0], direction)
        else:
            Messagebox.show_error("Selection error", "Select ONE item!", parent=self)


@doc_category("Storage widgets")
class ListBoxScrolled(ttk.Frame):
    """
    A scrollable version of :class:`tkclasswiz.storage.ListBoxObjects`.
    All the methods are the same as :class:`tkclasswiz.storage.ListBoxObjects`.
    """
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent)
        listbox = ListBoxObjects(self, *args, **kwargs)
        self.listbox = listbox

        listbox.pack(side="left", fill=tk.BOTH, expand=True)

        scrollbar = ttk.Scrollbar(self)
        scrollbar.pack(side=tk.RIGHT, fill=tk.BOTH)
        scrollbar.config(command=listbox.yview)

        listbox.config(yscrollcommand=scrollbar.set)

    def __getattr__(self, name: str):
        """
        Getter method that only get's called if the current
        implementation does not have the requested attribute.
        """
        return getattr(self.listbox, name)


@doc_category("Storage widgets")
class ComboBoxObjects(ttk.Combobox):
    """
    Modified version of :class:`tkinter.Combobox`, which:

    - Allows copy-paste and also sets the standard keyboard shortcuts for it
    - Overwrites the ``get`` method to return the original objects instead of text
    - Modifies the ``insert`` method to insert original objects instead of text
    """
    def __init__(self, *args, **kwargs):
        self._original_items = []
        super().__init__(*args, **kwargs)

    def save_to_clipboard(self):
        """
        Saves selected element to Clipboard.
        """
        GLOBAL.clipboard = self.get()

    def paste_from_clipboard(self):
        "Paste element(s) from clipboard."
        value = GLOBAL.clipboard
        if value is _NoClipBoard:
            return
        
        if value not in self._original_items:
            self.insert(tk.END, value)
        
        self.current(self._original_items.index(value))

    def get(self) -> Any:
        "Returns selected element"
        index = self.current()
        if isinstance(index, int) and index >= 0:
            return self._original_items[index]

        return super().get()

    def delete(self, index: int) -> None:
        """
        Removes the element at ``index``.
        """
        self["values"].pop(index)
        super().delete(index)
        self["values"] = self["values"]  # Update the text list, NOT a code mistake

    def insert(self, index: Union[int, str], element: Any) -> None:
        """
        Insert the ``element`` to the spot at ``index``.
        """
        if index == tk.END:
            self._original_items.append(element)
        else:
            self._original_items.insert(index, element)

        self["values"] = self._original_items

    def count(self) -> int:
        "Returns number of elements inside the ComboBox"
        return len(self._original_items)

    def __setitem__(self, key: str, value) -> None:
        if key == "values":
            self._original_items = list(value)
            value = [str(x)[:200] for x in value]

        return super().__setitem__(key, value)

    def __getitem__(self, key: str):
        if key == "values":
            return self._original_items

        return super().__getitem__(key)


@doc_category("Storage widgets")
class ComboEditFrame(ttk.Frame):
    """
    Frame, which combines :class:`tkclasswiz.storage.ComboBoxObjects` and an edit button.
    The edit button will open an object edit window / wizard and load in the old object (ObjectInfo) data.

    Parameters
    ----------------
    edit_method: Callable
        Function that opens the object edit window.
        This should either be :py:meth:`tkclasswiz.object_frame.window.ObjectEditWindow.open_object_edit_frame`,
        :py:meth:`tkclasswiz.object_frame.frame_base.NewObjectFrameBase.new_object_frame` or 
        a function that calls one of the previous 2 methods - this can be used to eg. load extra values into
        :class:`tkclasswiz.object_frame.frame_base.NewObjectFrameStruct`'s comboboxes by passing in the 
        ``additional_values`` mapping.

    values: list[Any]
        List of values that the combobox will have.
    master: Optional[:class:`tkinter.Widget`]
        The master widget.
    args
        Extra positional arguments to pass to :class:`tkinter.Frame``
    kwargs
        Extra keyword arguments to pass to :class:`tkinter.Frame``
    """
    def __init__(
        self,
        edit_method: Callable,
        values: List[Any] = [],
        master=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, master=master, **kwargs)
        combo = ComboBoxObjects(self)
        ttk.Button(self, text="Edit", command=self._edit).pack(side="right")
        combo.pack(side="left", fill=tk.X, expand=True)
        self.combo = combo
        self.edit_method = edit_method
        self.set_values(values)

    def set_values(self, values: List[Any] = []):
        "Sets the combobox values to ``values``."
        self.combo["values"] = values
        self.combo.current(0)

    def _edit(self):
        selection = self.combo.current()
        if selection >= 0:
            object_: ObjectInfo = self.combo.get()
            self.edit_method(
                object_.class_,
                self.combo,
                old_data=object_,
            )
        else:
            Messagebox.show_error("Empty list!", "Select atleast one item!")



from typing import TypeVar
from enum import Flag

from ..doc import doc_category
from ..deprecation import *
from ..utilities import *
from ..convert import *
from ..dpi import *

from ..storage import *

from .frame_base import *

import tkinter as tk
import tkinter.ttk as ttk


__all__ = ("NewObjectFrameFlag",)

T = TypeVar("T", ComboBoxObjects, ListBoxScrolled, ListBoxObjects)


@doc_category("Object frames")
class NewObjectFrameFlag(NewObjectFrameBase):
    """
    Frame for use inside the ObjectEditWindow that allows definition of enum flags.

    Parameters
    ------------
    class_: Flag
        The class we are defining for.
    return_widget: T
        The widget to insert the ObjectInfo into after saving.
    parent: TopLevel
        The parent window.
    old_data: Flag
        The old_data gui data.
    check_parameters: bool
        Check parameters (by creating the real object) upon saving.
        This is ignored if editing a function instead of a class.
    allow_save: bool
        If False, will open in read-only mode.
    """
    def __init__(
        self,
        class_: Flag,
        return_widget: T,
        parent: tk.Toplevel = None,
        old_data: Flag = None,
        check_parameters: bool = True,
        allow_save=True
    ):
        super().__init__(class_, return_widget, parent, old_data, check_parameters, allow_save)

        dpi_5 = dpi_scaled(5)
        dpi_10 = dpi_scaled(5)
        ttk.Label(self.frame_main, text="Current value").pack(anchor=tk.W)
        w = PyObjectScalar(self.frame_main)
        w.pack(fill=tk.X, pady=dpi_5)

        ttk.Separator(self.frame_main).pack(fill=tk.X, pady=dpi_10)

        ttk.Label(self.frame_main, text="Modify").pack(anchor=tk.W)
        combo_select = ComboBoxObjects(self.frame_main, width=max(map(len, map(str, list(class_)))))
        combo_select["values"] = list(class_)
        combo_select.pack(anchor=tk.W, pady=dpi_5)
        bnt_add_flag = ttk.Button(self.frame_main, text="Add flag", command=lambda: self._update_flag(combo_select.get(), True))
        bnt_add_flag.pack(anchor=tk.W)
        bnt_remove_flag = ttk.Button(self.frame_main, text="Remove flag", command=lambda: self._update_flag(combo_select.get(), False))
        bnt_remove_flag.pack(anchor=tk.W)

        self.storage_widget = w

        if old_data is not None:
            self.load(old_data)
        
        self.remember_gui_data()

    @gui_except()
    def _update_flag(self, flag: Flag, set: bool):
        if isinstance(flag, str):
            raise ValueError(
                f"No valid flag selected! Current selection: '{flag}'\n"
                f"Allowed values: {list(map(str, iter(self.class_)))}"
            )

        old_value = self.storage_widget.get()
        if old_value is None:
            old_value = flag

        if set:
            new_value = old_value | flag
        else:
            new_value = old_value & (~flag)

        self.storage_widget.set(new_value)

    def load(self, old_data: Flag):
        self.storage_widget.set(old_data)
        self.old_gui_data = old_data

    def get_gui_data(self):
        return self.storage_widget.get()

    def to_object(self) -> Flag:
        value = self.storage_widget.get()
        if isinstance(value, str):
            raise ValueError(f"Cannot directly input string '{value}'")

        if value is None:
            value = self.class_(0)

        return value



import tkinter.ttk as ttk
import tkinter as tk

from abc import ABCMeta, abstractmethod
from typing import Optional

from ..convert import ObjectInfo
from ..storage import *


class BaseToolTip(tk.Toplevel):
    """
    .. versionadded:: 1.2

    Used to display a nickname tooltip based on ``ObjectInfo.nickname`` attribute.
    It's triggered on ``enter_event`` after ``timeout_ms`` milliseconds and disappears on ``leave_event``.
    """
    __metaclass__ = ABCMeta

    def __init__(
        self,
        widget: tk.Widget,
        timeout_ms: int = 500,
    ):
        super().__init__(widget)
        ttk.Style().configure(
            style="tooltip.TLabel",  # ttkbootstrap compatibility
            background="white",
        )
        self.label = ttk.Label(self, style="tooltip.TLabel", wraplength=1000)
        self.schedule_id = None
        self._widget = widget
        self.timeout_ms = timeout_ms
        self.label.pack()
        self.pack_propagate(True)
        self._hide_tooltip()
        self.overrideredirect(True)
        self.attributes('-topmost', True)

    def _schedule(self, event: tk.Event):
        if not (value := self._get_value()):
            return

        self.label.config(text=str(value)[:3000])
        if self.timeout_ms:
            self.schedule_id = self.after(self.timeout_ms, lambda: self._show_tooltip(event))
        else:
            self.after_idle(lambda: self._show_tooltip(event))

    def _cancel_schedule(self, event: tk.Event):
        if self.schedule_id:
            self.after_cancel(self.schedule_id)
            self.schedule_id = None

        self._hide_tooltip()

    def _show_tooltip(self, event: tk.Event):
        self.geometry("")
        self.deiconify()
        self._update_pos(event)
        self._widget.bind("<Motion>", self._update_pos)

    def _update_pos(self, event: tk.Event):
        geo = self.geometry().split('+')[0]
        x, y = self.winfo_pointerxy()
        self.geometry(f'{geo}+{x + 10}+{y + 10}')

    def _hide_tooltip(self):
        self.withdraw()
        

    @abstractmethod
    def _get_value(self) -> Optional[ObjectInfo]:
        pass



class ListboxTooltip(BaseToolTip):
    def __init__(self, widget: tk.Widget, timeout_ms: int = 500):
        if isinstance(widget, ListBoxScrolled):
            widget = widget.listbox

        super().__init__(widget, timeout_ms)
        self._widget.bind("<<ListboxSelect>>", self._schedule)
        self._widget.bind("<Leave>", self._cancel_schedule)
        self.start_y = 0

    def _get_value(self):
        value = self._widget.get()
        selection = self._widget.curselection()
        if len(selection) != 1:
            return

        value = value[self._widget.current()]
        return str(value)

    def _show_tooltip(self, event: tk.Event):
        self.start_y = self.winfo_pointery()
        super()._show_tooltip(event)

    def _update_pos(self, event: tk.Event):
        super()._update_pos(event)
        if abs(self.start_y - self.winfo_pointery()) > 10:
            self._cancel_schedule(event)


class ComboboxTooltip(BaseToolTip):
    def __init__(self, widget: tk.Widget, timeout_ms: int = 500):
        super().__init__(widget, timeout_ms)
        self._widget.bind("<Enter>", self._schedule)
        self._widget.bind("<Leave>", self._cancel_schedule)

    def _get_value(self):
        value = self._widget.get()
        return str(value)
